import copy

# We implement Bayesian networks.
# For simplificity, all probability variables are Booleans.
#
# The networks are represented as Python dictionaries. Below is a sketch
# of the network for the burglary example from the lecture notes.
# It states, for example, that P(alarm | burglary, not earthquake) = 0.94.
# It follows that $P(not alarm | burglary, not earthquake) = 1-0.94.
#
# example_network = {
#     'Burglary': {'name': 'Burglary', 'parents': [], 'probabilities': {(): 0.001}},
#     'Earthquake': {'name': 'Earthquake', 'parents': [], 'probabilities': {(): 0.002}},
#     'Alarm': {
#         'name': 'Alarm',
#         'parents': ['Burglary', 'Earthquake'],
#         'probabilities': {
#             (True, True): 0.95,
#             (True, False): 0.94,
#             (False, True): 0.29,
#             (False, False): 0.001}
#         },
#     'JohnCalls': {'name': 'JohnCalls', 'parents': ['Alarm'], 'probabilities': {(True,): 0.9, (False,): 0.05}},
#     'MaryCalls': {'name': 'MaryCalls', 'parents': ['Alarm'], 'probabilities': {(True,): 0.7, (False,): 0.01}}
# }
#
# Queries consist of the network (as above), a single query variable, and an atomic event for the evidence.
# The latter is a dictionary that gives for the every evidence variable a value.
#
# Example query: query(example_network, 'Burglary', {'MaryCalls':True, 'JohnCalls':True})



def combiner(old_comb, new):
    if len(old_comb) == 0:
        return new
    else:
        result = []
        for i in old_comb:
            for t in new:
                new_dict = {**i , **t}
                result.append(new_dict)
    return result


def dict_creator(name, possible_vales):
    result = []
    for v in possible_vales:
        dict = {name: v}
        result.append(dict)
    return result


def generate_all_sequences(network):
    possible_vales = [True, False]
    value_per_variable =[]

    all_seq = []

    for key in network:
        value_per_variable.append(dict_creator(key, possible_vales))

    for r in value_per_variable:
        all_seq = combiner(all_seq, r)

    return all_seq

def filter_seq(all_seq, given):
    result = []
    for seq in all_seq:
        is_accepted = True
        for k, v in given.items():
            if seq[k] != v:
                is_accepted = False
                break
        if is_accepted:
            result.append(seq)
    return result

def parse_prob(v, required_value):
    if len(v['parents']) == 0:
        return v['probabilities'][()]
    elif len(v['parents']) == 1:
        return v['probabilities'][(required_value,)]
    else:
        return v['probabilities'][required_value]


def seq_prob_cal(network, seq):
    prob = 1
    for leaf, value in seq.items():
        current_prob = 1
        v = network[leaf]
        if len(v["parents"]) == 1:
            current_prob = parse_prob(v, seq[v["parents"][0]])
        elif len(v["parents"]) == 0:
            current_prob = parse_prob(v, True)
        else:
            cond = []
            for parent in v["parents"]:
                cond.append(seq[parent])
            current_prob = parse_prob(v, tuple(cond))


        if value:
            prob *= current_prob
        else:
            prob *= (1-current_prob)
    return prob

def normalization(q):
    prob_sum = 0
    for key, value in q.items():
        prob_sum += value
    alpha = 1 / prob_sum

    new_q = {}
    for key, value in q.items():
        new_q[key] = alpha * value

    return  new_q


##
# The goal is to calculate P(node = true | evidence) the evidence here is 
# mary call or john call so we need to calculate the probability of burlary given mary call = true 
# and john call = true for example
# all the other variables are hidden variables (earthquake and alarm) so we will have to marginalize over them
# lets assume the queey is p(node = true | mary call = true, john call = true)
# the function generate_all_sequences the possible combination of the variables for example node = true, mary call = true, john call = true, burglary = true, earthquake = true, alarm = true and so on 
# the function filter_seq filters the sequences that are not consistent with the evidence for example if the evidence is mary call = true, john call = true then 
# the sequence node = true, mary call = false, john call = true is not consistent with the evidence so it will be filtered out
# after the sequences have been filtered out we can calculate the probability of each sequence using the function seq_prob_cal
# the last step is to normalize the probabilities so that they sum up to 1
## 
def query(network, node, evidence):
    q = {True: 0, False: 0}
    for key, value in q.items():
        given = {}
        given[node] = key
        given.update(evidence)

        all_seq = generate_all_sequences(network)
        filtered_seq = filter_seq(all_seq, given)
        prob_sum = 0
        for seq in filtered_seq:
            prob = seq_prob_cal(network, seq)
            prob_sum += prob

        q[key] = prob_sum

    new_q = normalization(q)
    return new_q[True]


